// -Tile Engine Demo-
// Written by J. Farrell

// This program is a very simple tile engine, using two-layer maps. It
// supports tile animations as well, but currently does not have any
// characters. It has eight-way scrolling, which the user can control
// via the arrow keys. Press Escape to quit the demo.

#include "Tile Engine Demo.h"
#include "resource.h"

// To compile this program, remember what you need to do to set up
// Visual C++ for DirectX. Make sure the DirectX SDK directories
// for libraries and headers are FIRST on their repective lists
// (under Tools menu, go to "Options..."). Then create a new project
// using this and Script1.rc as the only source files, and remember
// to link the ddraw.lib and dxguid.lib libraries to your project!

// G L O B A L S /////////////////////////////////////////////////////
// Windows
HINSTANCE hMainInstance;
HWND hMainWindow;

// DirectDraw
LPDIRECTDRAW7 lpdd7 = NULL;
LPDIRECTDRAWSURFACE7 lpddsPrimary = NULL;
LPDIRECTDRAWSURFACE7 lpddsBack = NULL;
LPDIRECTDRAWSURFACE7 lpddsTileset = NULL;
LPDIRECTDRAWCLIPPER  lpddClipper = NULL;
DDSURFACEDESC2 ddsd;

// tile engine
BYTE byMap[300][300][2];         // the actual tiles
MAPDATA mapdata;                 // the other map data
TILE tileData[NUM_TILES];
LPTILE lptile[NUM_TILES];

//////////////////////////////////////////////////////////////////////
// WINMAIN
// Program entry point. Handles the window setup and the contains the
// program's main loop.
//////////////////////////////////////////////////////////////////////
int APIENTRY WinMain(HINSTANCE hinstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	WNDCLASSEX winclass; // our window class
	HWND hwnd;           // handle to a window
	MSG msg;             // a message variable
	int x;

	// fill out the winclass structure
	winclass.cbSize =        sizeof(WNDCLASSEX);
	winclass.style =         CS_DBLCLKS | CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
	winclass.lpfnWndProc =   WindowProc;        
	winclass.cbClsExtra =    0;                  // extra class info
	winclass.cbWndExtra =    0;                  // extra window info
	winclass.hInstance =     hinstance;	         
	winclass.hIcon =         LoadIcon(NULL, IDI_WINLOGO);
	winclass.hCursor =       LoadCursor(NULL, IDC_ARROW);
	winclass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	winclass.lpszMenuName =  NULL;
	winclass.lpszClassName = "Tile Engine Class";  
	winclass.hIconSm =       LoadIcon(NULL, IDI_WINLOGO);
	
	// register class
	if (!RegisterClassEx(&winclass))
		return(0);
	
	// create a window
	if (!(hwnd = CreateWindowEx(NULL,            // extended style
			     "Tile Engine Class",            // class identifier
				 "Basic Tile Engine",            // window title
				 WS_POPUP | WS_VISIBLE,          // window style
				 0, 0,							 // initial position
				 640, 480,						 // initial size
				 NULL,                           // handle to parent
				 NULL,                           // handle to menu
				 hinstance,                      // app handle
				 NULL)))                         // extra parameters
		return(0);

	// set globals
	hMainWindow = hwnd;
	hMainInstance = hinstance;

	// initialize, test for success
	if (Initialize())
	{
		// main program loop
		while (TRUE)
		{
			// get the time
			DWORD dwStart = GetTickCount();

			// check the message queue
			if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
			{
				if (msg.message == WM_QUIT) // exit loop on quit
					break;
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}

			// check user input
			if (KEYSTATE(VK_UP) && (mapdata.yCamera > 4))
				mapdata.yCamera -= 5;
			if (KEYSTATE(VK_DOWN) && (mapdata.yCamera < (mapdata.yMaxCamera - 4)))
				mapdata.yCamera += 5;
			if (KEYSTATE(VK_LEFT) && (mapdata.xCamera > 4))
				mapdata.xCamera -= 5;
			if (KEYSTATE(VK_RIGHT) && (mapdata.xCamera < (mapdata.xMaxCamera - 4)))
				mapdata.xCamera += 5;

			// draw and show frame
			Render_Map();
			lpddsPrimary->Flip(NULL, DDFLIP_WAIT);

			// check escape key
			if (KEYSTATE(VK_ESCAPE))
				PostQuitMessage(0);

			// advance animations
			for (x=0; x<NUM_TILES; x++)
			{
				// update each frame counter
				if (lptile[x]->nAnimSpeed > 0)
				{
					// advance frame counter
					lptile[x]->uiFrames++;

					// check if we're finished
					if (lptile[x]->uiFrames == (UINT)lptile[x]->nAnimSpeed)
					{
						lptile[x]->uiFrames = 0;
						lptile[x] = lptile[x]->lpNext;
					}
				}
			}

			// lock to 30 FPS (33ms per frame)
			while ((GetTickCount() - dwStart) < 33);
		}
	}
	else
		MessageBox(hMainWindow, "Initialization failed!", "Error", MB_OK | MB_ICONEXCLAMATION);
	
	// shut everything down
	Shutdown();

	// give Windows the shutdown code
	return(msg.wParam);
}

LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg,
							WPARAM wparam, LPARAM lparam)
{
    // lazy man's message handler
	return(DefWindowProc(hwnd, msg, wparam, lparam));
}

///////////////////////////////////////////////////////////////
// INIT_MAP
// Loads the map, fills out the map data structure, and sets up
// the tile data structures. Be sure to read my comments here.
///////////////////////////////////////////////////////////////
int Init_Map(void)
{
	FILE *pfile;
	int x, y;

	// first open the map
	if ((pfile = fopen("testmap.dat", "rb")) == NULL)
		return(FALSE);

	// read the map dimensions
	mapdata.xMax = getc(pfile);
	mapdata.yMax = getc(pfile);

	// now read the actual map data and close the file
	for (x=0; x<=mapdata.xMax; x++)
	{
		for (y=0; y<=mapdata.yMax; y++)
		{
			byMap[x][y][0] = getc(pfile);
			byMap[x][y][1] = getc(pfile);
		}
	}
	fclose(pfile);

	// compute the maximum camera values
	mapdata.xMaxCamera = (mapdata.xMax - 19) * 32;
	if (mapdata.xMaxCamera < 0)
		mapdata.xMaxCamera = 0;
	mapdata.yMaxCamera = (mapdata.yMax - 14) * 32;
	if (mapdata.yMaxCamera < 0)
	    mapdata.yMaxCamera = 0;

	// set the camera at the center of the map
	mapdata.xCamera = mapdata.xMaxCamera / 2;
	mapdata.yCamera = mapdata.yMaxCamera / 2;

	// IMPORTANT - set the tile pointers initially pointing
	// at their corresponding tile structures
	for (x=0; x<NUM_TILES; x++)
		lptile[x] = &tileData[x];

	// Now we're going to set up the tile structures, but this is not normally
	// how it would be done. For an actual game tileset, which can get quite
	// large, hard-coding all this information is a very bad idea. What you'd
	// actually want to do is to create a data file that contains all of this
	// information, so the game can just open the data file and read in what
	// it needs. Because this example tileset is so small, and to keep things
	// easy to understand (i.e. not having tons of data files), I'm just going
	// to fill out the structures here.
	for (x=0; x<NUM_TILES; x++)
	{
		// set all the RECTs on the surface
		RECT rcTemp = {0, x*32, 32, x*32+32};
		tileData[x].rcLocation = rcTemp;

		// reset the frame counter
		tileData[x].uiFrames = 0;
	}

	// set up the animation chain for the water
	tileData[4].lpNext = &tileData[5]; tileData[4].nAnimSpeed = 30;
	tileData[5].lpNext = &tileData[6]; tileData[5].nAnimSpeed = 3;
	tileData[6].lpNext = &tileData[7]; tileData[6].nAnimSpeed = 3;
	tileData[7].lpNext = &tileData[4]; tileData[7].nAnimSpeed = 3;
	
	// bWalkOK and dwFlags are not important since there are no characters used
	// in this demo, so we don't even bother to set them here.

	// return success
	return(TRUE);
}

///////////////////////////////////////////////////////////////
// INITIALIZE
// Creates all the DirectX objects we'll need for this demo,
// sets the screen resolution to 640x480x16.
///////////////////////////////////////////////////////////////
int Initialize(void)
{
	// create the DirectDraw7 interface
	if (FAILED(DirectDrawCreateEx(NULL, (void**)&lpdd7, IID_IDirectDraw7, NULL)))
		return(FALSE);

	// set cooperative level to fullscreen
	if (FAILED(lpdd7->SetCooperativeLevel(hMainWindow, DDSCL_FULLSCREEN | 
		                                               DDSCL_EXCLUSIVE |
													   DDSCL_ALLOWREBOOT)))
		return(FALSE);

	// hide the mouse
	ShowCursor(FALSE);

	// set 640x480x16 resolution
	if (FAILED(lpdd7->SetDisplayMode(640, 480, 16, 0, 0)))
		return(FALSE);

	// create our primary surface, with a back buffer
	INIT_DXSTRUCT(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
						  DDSCAPS_COMPLEX |
						  DDSCAPS_FLIP;
	ddsd.dwBackBufferCount = 1;
	if (FAILED(lpdd7->CreateSurface(&ddsd, &lpddsPrimary, NULL)))
		return(FALSE);

	// get attached back buffer
	ddsd.ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;
	if (FAILED(lpddsPrimary->GetAttachedSurface(&ddsd.ddsCaps, &lpddsBack)))
		return(FALSE);

	// create the clipper
	if (FAILED(lpdd7->CreateClipper(0, &lpddClipper, NULL)))
		return(FALSE);

	// create and set the clip list
	LPRGNDATA lpClipList = (LPRGNDATA)malloc(sizeof(RGNDATAHEADER) +
										     sizeof(RECT));
	RECT rcBoundary = {0, 0, 640, 480};
	memcpy(lpClipList->Buffer, &rcBoundary, sizeof(RECT)); // set list
	lpClipList->rdh.dwSize = sizeof(RGNDATAHEADER);        // structure size
	lpClipList->rdh.iType = RDH_RECTANGLES;                // bound type
	lpClipList->rdh.nCount = 1;                            // num of bounds
	lpClipList->rdh.nRgnSize = sizeof(RECT);               // size of Buffer
	lpClipList->rdh.rcBound = rcBoundary;                  // bounding RECT
	if (FAILED(lpddClipper->SetClipList(lpClipList, 0)))
	{
		// always free the mallocs!
		free(lpClipList);
		return(FALSE);
	}

	// attach clipper to back buffer
	if (FAILED(lpddsBack->SetClipper(lpddClipper)))
	{
		free(lpClipList);
		return(FALSE);
	}
	
	// free the memory
	free(lpClipList);

	// create the tileset surface (dimensions #defined in header file)
	// be sure to set a color key of black only
	INIT_DXSTRUCT(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_CKSRCBLT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	ddsd.dwWidth = TILESET_WIDTH;
	ddsd.dwHeight = TILESET_HEIGHT;
	ddsd.ddckCKSrcBlt.dwColorSpaceLowValue  = 0;
	ddsd.ddckCKSrcBlt.dwColorSpaceHighValue = 0;
	if (FAILED(lpdd7->CreateSurface(&ddsd, &lpddsTileset, NULL)))
		return(FALSE);

	// load the image onto the surface
	if (!Load_Bitmap(MAKEINTRESOURCE(IDB_BITMAP1), 0, 0, lpddsTileset))
		return(FALSE);

	// now set up the map
	if (!Init_Map())
		return(FALSE);

	// all set!
	return(TRUE);
}

//////////////////////////////////////////////////////////////////////
// LOAD_BITMAP
// Loads a bitmap resource onto the specified DirectDraw surface.
//////////////////////////////////////////////////////////////////////
int Load_Bitmap(char *lpszName, int x, int y, LPDIRECTDRAWSURFACE7 lpdds)
{
	HBITMAP hbitmap = NULL;
	BITMAP bmp;
	HDC hdcImage, hdcSurface;

	// load bitmap resource
	if (!(hbitmap = (HBITMAP)LoadImage(hMainInstance, lpszName, 
		                               IMAGE_BITMAP, 0, 0, 
					     			   LR_CREATEDIBSECTION)))
		return(FALSE);

	// create device context for image
	if (!(hdcImage = CreateCompatibleDC(NULL)))
		return(FALSE);

	// select bitmap into the DC
	if (!(SelectObject(hdcImage, hbitmap)))
		return(FALSE);

	// get surface device context
	if (FAILED(lpdds->GetDC(&hdcSurface)))
		return(FALSE);

	// get image dimensions
	GetObject(hbitmap, sizeof(BITMAP), &bmp);
    int dx = bmp.bmWidth;
    int dy = bmp.bmHeight;

	// copy image to surface
	if (!(BitBlt(hdcSurface, x, y, dx, dy, hdcImage, 0, 0, SRCCOPY)))
		return(FALSE);
	
	// get rid of all that stuff we used
	lpdds->ReleaseDC(hdcSurface);
	DeleteDC(hdcImage);
	DeleteObject(hbitmap);

	return(TRUE);
}

//////////////////////////////////////////////////////////////////////
// RENDER_MAP
// Draws the map on the back buffer.
//////////////////////////////////////////////////////////////////////
void Render_Map(void)
{
	BYTE byTile;

	// set original destination RECT for first tile -- aligned with the
	// upper-left corner of the screen
	RECT rcDest = {0, 0, 32, 32};

	// find default tile range -- divide camera coordinates by 32 and
	// use the default of 21 tiles across and 16 tiles down
	int xStart = mapdata.xCamera >> 5;
	int yStart = mapdata.yCamera >> 5;
	int xEnd = xStart + 20;
	int yEnd = yStart + 15;
	int x, y;

	// now check if the camera coordinates are divisible by 32
	x = mapdata.xCamera & 0x0000001F;
	y = mapdata.yCamera & 0x0000001F;

	if (!x)
	{
		// if xCamera is divisible by 32, use only 20 tiles across
		xEnd--;
	}
	else
	{
		// otherwise move destination RECT to the left to clip the first column of tiles
		rcDest.left-=x;
		rcDest.right-=x;
	}

	if (!y)
	{
		// if yCamera is divisible by 32, use only 15 tiles down
		yEnd--;
	}
	else
	{
		// otherwise move destination RECT up to clip the first row of tiles
		rcDest.top-=y;
		rcDest.bottom-=y;
	}

	// finally make sure we're not exceeding map limits
	if (xEnd > mapdata.xMax)
		xEnd = mapdata.xMax;
	if (yEnd > mapdata.yMax)
		yEnd = mapdata.yMax;

	// store original rcDest RECT
	RECT rcTemp;
	rcTemp = rcDest;

	// plot the first layer
	for (x=xStart; x<=xEnd; x++)
	{
		for (y=yStart; y<=yEnd; y++)
		{
			// blit the tile
			byTile = byMap[x][y][0];
			lpddsBack->Blt(&rcDest, lpddsTileset, &lptile[byTile]->rcLocation, DDBLT_WAIT, NULL);

			// advance rcDest RECT
			rcDest.bottom += 32;
			rcDest.top += 32;
		}

		// reset rcDest RECT to top of next column
		rcDest.left += 32;
		rcDest.right += 32;
		rcDest.bottom -= ((yEnd - yStart + 1) << 5);
		rcDest.top -= ((yEnd - yStart + 1) << 5);
	}

	// restore the destination RECT
	rcDest = rcTemp;

	// plot the second layer
	for (x=xStart; x<=xEnd; x++)
	{
		for (y=yStart; y<=yEnd; y++)
		{
			// blit the tile if necessary
			byTile = byMap[x][y][1];
			if (byTile > 0)
				lpddsBack->Blt(&rcDest, lpddsTileset, &lptile[byTile]->rcLocation, DDBLT_WAIT | DDBLT_KEYSRC, NULL);

			// advance rcDest RECT
			rcDest.bottom += 32;
			rcDest.top += 32;
		}

		// reset rcDest RECT to top of next column
		rcDest.left += 32;
		rcDest.right += 32;
		rcDest.bottom -= ((yEnd - yStart + 1) << 5);
		rcDest.top -= ((yEnd - yStart + 1) << 5);
	}
}

///////////////////////////////////////////////////////////////
// SHUTDOWN
// Release the objects we created.
///////////////////////////////////////////////////////////////
void Shutdown(void)
{
	if (lpddsTileset)
	{
		// a pox on the tileset surface!
		lpddsTileset->Release();
		lpddsTileset = NULL;
	}

	if (lpddClipper)
	{
		// death to the clipper!
		lpddClipper->Release();
		lpddClipper = NULL;
	}

	if (lpddsBack)
	{
		// destroy the back buffer!
		lpddsBack->Release();
		lpddsBack = NULL;
	}

	if (lpddsPrimary)
	{
		// a plague upon the primary surface!
		lpddsPrimary->Release();
		lpddsPrimary = NULL;
	}

	if (lpdd7)
	{
		// blow away DirectDraw!
		lpdd7->Release();
		lpdd7 = NULL;
	}

	// dismember the window!
	CloseWindow(hMainWindow);
}
